@@ -37,6 +37,7 @@ import java.io.File;
 import java.io.IOException;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collection;
 import java.util.Collections;
 import java.util.HashSet;
 import java.util.Iterator;
@@ -51,7 +52,7 @@ import java.util.Set;
  * matching entries, etc.
  *
  * @author jdcasey
- * @version $Id: FileSetManager.java 597982 2007-11-25 13:05:59Z vsiveton $
+ * @version $Id: FileSetManager.java 694312 2008-09-11 15:46:15Z bentmann $
  */
 public class FileSetManager
 {
@@ -135,8 +136,6 @@ public class FileSetManager
 
     /**
      * Create a new manager instance with an empty messages. Verbose flag is set to false.
-     *
-     * @param log The mojo log instance
      */
     public FileSetManager()
     {
@@ -299,14 +298,38 @@ public class FileSetManager
 
             if ( file.exists() )
             {
-                if ( file.isDirectory() && ( fileSet.isFollowSymlinks() || !isSymlink( file ) ) )
+                if ( file.isDirectory() )
                 {
-                    if ( verbose && messages != null )
+                    if ( fileSet.isFollowSymlinks() || !isSymlink( file ) )
                     {
-                        messages.addInfoMessage( "Deleting directory: " + file ).flush();
+                        if ( verbose && messages != null )
+                        {
+                            messages.addInfoMessage( "Deleting directory: " + file ).flush();
+                        }
+    
+                        removeDir( file, fileSet.isFollowSymlinks(), throwsError, warnMessages );
                     }
+                    else
+                    { // delete a symlink to a directory without follow
+                        if ( verbose && messages != null )
+                        {
+                            messages.addInfoMessage( "Deleting symlink to directory: " + file ).flush();
+                        }
+    
+                        if ( !file.delete() )
+                        {
+                            String message = "Unable to delete symlink " + file.getAbsolutePath();
+                            if ( throwsError )
+                            {
+                                throw new IOException( message );
+                            }
 
-                    removeDir( file, fileSet.isFollowSymlinks(), throwsError, warnMessages );
+                            if ( !warnMessages.contains( message ) )
+                            {
+                                warnMessages.add( message );
+                            }
+                        }
+                    }
                 }
                 else
                 {
@@ -347,17 +370,25 @@ public class FileSetManager
     private boolean isSymlink( File file )
         throws IOException
     {
-        File parent = file.getParentFile();
-        File canonicalFile = file.getCanonicalFile();
-
+        File fileInCanonicalParent = null;
+        File parentDir = file.getParentFile();
+        if ( parentDir == null )
+        {
+            fileInCanonicalParent = file;
+        }
+        else
+        {
+            fileInCanonicalParent = new File( parentDir.getCanonicalPath(), file.getName() );
+        }
         if ( messages != null && messages.isDebugEnabled() )
         {
-            messages.addDebugMessage( "Checking for symlink:\nParent file's canonical path: "
-                + parent.getCanonicalPath() + "\nMy canonical path: " + canonicalFile.getPath() ).flush();
+            messages.addDebugMessage(
+                                      "Checking for symlink:\nFile's canonical path: "
+                                          + fileInCanonicalParent.getCanonicalPath()
+                                          + "\nFile's absolute path with canonical parent: "
+                                          + fileInCanonicalParent.getPath() ).flush();
         }
-        return parent != null
-            && ( !canonicalFile.getName().equals( file.getName() ) || !canonicalFile.getPath()
-                .startsWith( parent.getCanonicalPath() ) );
+        return !fileInCanonicalParent.getCanonicalFile().equals( fileInCanonicalParent.getAbsoluteFile() );
     }
 
     private Set findDeletablePaths( FileSet fileSet )
@@ -382,12 +413,9 @@ public class FileSetManager
             return Collections.EMPTY_SET;
         }
 
-        String[] includedDirs = scanner.getIncludedDirectories();
-        String[] excludedDirs = scanner.getExcludedDirectories();
-
-        Set includes = new HashSet( Arrays.asList( includedDirs ) );
-        List excludes = new ArrayList( Arrays.asList( excludedDirs ) );
-        List linksForDeletion = new ArrayList();
+        Set includes = new HashSet( Arrays.asList( scanner.getIncludedDirectories() ) );
+        Collection excludes = new ArrayList( Arrays.asList( scanner.getExcludedDirectories() ) );
+        Collection linksForDeletion = new ArrayList();
 
         if ( !fileSet.isFollowSymlinks() )
         {
@@ -397,7 +425,7 @@ public class FileSetManager
                     .addInfoMessage( "Adding symbolic link dirs which were previously excluded to the list being deleted." ).flush();
             }
 
-            // we need to see which entries were excluded because they're symlinks...
+            // we need to see which entries were only excluded because they're symlinks...
             scanner.setFollowSymlinks( true );
             scanner.scan();
 
@@ -407,45 +435,20 @@ public class FileSetManager
                 messages.addDebugMessage( "Marked for preserve (with followSymlinks == false): " + excludes ).flush();
             }
 
-            List notSymlinks = Arrays.asList( scanner.getIncludedDirectories() );
+            List includedDirsAndSymlinks = Arrays.asList( scanner.getIncludedDirectories() );
 
             linksForDeletion.addAll( excludes );
-            linksForDeletion.retainAll( notSymlinks );
+            linksForDeletion.retainAll( includedDirsAndSymlinks );
 
             if ( messages != null && messages.isDebugEnabled() )
             {
                 messages.addDebugMessage( "Symlinks marked for deletion (originally mismarked): " + linksForDeletion ).flush();
             }
 
-            excludes.removeAll( notSymlinks );
+            excludes.removeAll( includedDirsAndSymlinks );
         }
 
-        for ( int i = 0; i < excludedDirs.length; i++ )
-        {
-            String path = excludedDirs[i];
-
-            File excluded = new File( path );
-
-            String parentPath = excluded.getParent();
-
-            while ( parentPath != null )
-            {
-                if ( messages != null && messages.isDebugEnabled() )
-                {
-                    messages.addDebugMessage( "Verifying path: " + parentPath
-                        + " is not present; contains file which is excluded." ).flush();
-                }
-
-                boolean removed = includes.remove( parentPath );
-
-                if ( removed && messages != null && messages.isDebugEnabled() )
-                {
-                    messages.addDebugMessage( "Path: " + parentPath + " was removed from delete list." ).flush();
-                }
-
-                parentPath = new File( parentPath ).getParent();
-            }
-        }
+        excludeParentDirectoriesOfExcludedPaths( excludes, includes );
 
         includes.addAll( linksForDeletion );
 
@@ -466,13 +469,10 @@ public class FileSetManager
             return deletableDirectories;
         }
 
-        String[] includedFiles = scanner.getIncludedFiles();
-        String[] excludedFiles = scanner.getExcludedFiles();
-
         Set includes = deletableDirectories;
-        includes.addAll( Arrays.asList( includedFiles ) );
-        List excludes = new ArrayList( Arrays.asList( excludedFiles ) );
-        List linksForDeletion = new ArrayList();
+        includes.addAll( Arrays.asList( scanner.getIncludedFiles() ) );
+        Collection excludes = new ArrayList( Arrays.asList( scanner.getExcludedFiles() ) );
+        Collection linksForDeletion = new ArrayList();
 
         if ( !fileSet.isFollowSymlinks() )
         {
@@ -482,7 +482,7 @@ public class FileSetManager
                     .addInfoMessage( "Adding symbolic link files which were previously excluded to the list being deleted." ).flush();
             }
 
-            // we need to see which entries were excluded because they're symlinks...
+            // we need to see which entries were only excluded because they're symlinks...
             scanner.setFollowSymlinks( true );
             scanner.scan();
 
@@ -492,59 +492,77 @@ public class FileSetManager
                 messages.addDebugMessage( "Marked for preserve (with followSymlinks == false): " + excludes ).flush();
             }
 
-            List notSymlinks = Arrays.asList( scanner.getExcludedFiles() );
+            List includedFilesAndSymlinks = Arrays.asList( scanner.getIncludedFiles() );
 
             linksForDeletion.addAll( excludes );
-            linksForDeletion.retainAll( notSymlinks );
+            linksForDeletion.retainAll( includedFilesAndSymlinks );
 
             if ( messages != null && messages.isDebugEnabled() )
             {
                 messages.addDebugMessage( "Symlinks marked for deletion (originally mismarked): " + linksForDeletion ).flush();
             }
 
-            excludes.removeAll( notSymlinks );
+            excludes.removeAll( includedFilesAndSymlinks );
         }
 
-        for ( int i = 0; i < excludedFiles.length; i++ )
+        excludeParentDirectoriesOfExcludedPaths( excludes, includes );
+
+        includes.addAll( linksForDeletion );
+
+        return includes;
+    }
+
+    /**
+     * Removes all parent directories of the already excluded files/directories from the given set of deletable
+     * directories. I.e. if "subdir/excluded.txt" should not be deleted, "subdir" should be excluded from deletion, too.
+     * 
+     * @param excludedPaths The relative paths of the files/directories which are excluded from deletion, must not be
+     *            <code>null</code>.
+     * @param deletablePaths The relative paths to files/directories which are scheduled for deletion, must not be
+     *            <code>null</code>.
+     */
+    private void excludeParentDirectoriesOfExcludedPaths( Collection excludedPaths, Set deletablePaths )
+    {
+        for ( Iterator it = excludedPaths.iterator(); it.hasNext(); )
         {
-            String path = excludedFiles[i];
+            String path = (String) it.next();
 
-            File excluded = new File( path );
-
-            String parentPath = excluded.getParent();
+            String parentPath = new File( path ).getParent();
 
             while ( parentPath != null )
             {
                 if ( messages != null && messages.isDebugEnabled() )
                 {
-                    messages.addDebugMessage( "Verifying path: " + parentPath
+                    messages.addDebugMessage( "Verifying path " + parentPath
                         + " is not present; contains file which is excluded." ).flush();
                 }
 
-                boolean removed = includes.remove( parentPath );
+                boolean removed = deletablePaths.remove( parentPath );
 
                 if ( removed && messages != null && messages.isDebugEnabled() )
                 {
-                    messages.addDebugMessage( "Path: " + parentPath + " was removed from delete list." ).flush();
+                    messages.addDebugMessage( "Path " + parentPath + " was removed from delete list." ).flush();
                 }
 
                 parentPath = new File( parentPath ).getParent();
             }
         }
 
-        includes.addAll( linksForDeletion );
+        if ( !excludedPaths.isEmpty() )
+        {
+            if ( messages != null && messages.isDebugEnabled() )
+            {
+                messages.addDebugMessage( "Verifying path " + "."
+                    + " is not present; contains file which is excluded." ).flush();
+            }
 
-        // for ( Iterator it = includes.iterator(); it.hasNext(); )
-        // {
-        // String path = (String) it.next();
-        //
-        // if ( includes.contains( new File( path ).getParent() ) )
-        // {
-        // it.remove();
-        // }
-        // }
+            boolean removed = deletablePaths.remove( "" );
 
-        return includes;
+            if ( removed && messages != null && messages.isDebugEnabled() )
+            {
+                messages.addDebugMessage( "Path " + "." + " was removed from delete list." ).flush();
+            }
+        }
     }
 
     /**
